#include <cstdio>

#include "casegen.hh"

void Emit(const std::string& s, const std::string& p0 = "")
{
    std::printf("\t%s", s.c_str());
    if(!p0.empty())std::printf(" %s", p0.c_str());
    std::printf("\n");
}
void EmitImmed(const std::string& s, int p0)
{
    char Buf[64]; std::sprintf(Buf, "'%c'", p0);
    Emit(s, Buf);
}
const std::string GenLabel()
{
    static unsigned counter = 1;
    char Buf[64];
    sprintf(Buf, "L%u", counter++);
    return Buf;
}
void EmitLabel(const std::string& label)
{
    Emit(label + ":");
}

class CaseHandler: public CaseGenerator
{
    bool         LastWasBra;
    bool         ChainCompare;
    CaseValue    LastCompare;
    std::string  LastCompareType;
private:
    void EmitCompareJump
       (CaseValue value,
        const std::string& target,
        const char *comparetype)
    {
        if(!ChainCompare || value != LastCompare)
        {
            EmitImmed("cmp", value),
            LastCompare  = value;
            ChainCompare = true;
        }
        Emit(comparetype, target.c_str());
        LastWasBra      = false;
        LastCompareType = comparetype;
    }

public:
    CaseHandler()
    : LastWasBra(false), ChainCompare(false)
    {
    }
    
    virtual CaseValue GetMinValue()    const { return 0;   }
    virtual CaseValue GetMaxValue()    const { return 255; }
    virtual CaseValue GetDefaultCase() const { return -1;  }

    virtual void EmitJump(const std::string& target)
    {
        if(LastWasBra) return;
        Emit("bra", target.c_str());
        LastWasBra   = true;
        ChainCompare = false;
    }
    virtual void EmitCompareEQ(CaseValue value, const std::string& target)
    {
        EmitCompareJump(value, target, "beq");
    }
    virtual void EmitCompareNE(CaseValue value, const std::string& target)
    {
        EmitCompareJump(value, target, "bne");
    }
    virtual void EmitCompareLT(CaseValue value, const std::string& target)
    {
        if(value == GetMinValue())
         { /* nothing */ }
     else if(value == GetMinValue()+1
        && (!ChainCompare || value != LastCompare))
         EmitCompareEQ(value-1, target);
     else
         EmitCompareJump(value, target, "bcc");
    }
    virtual void EmitCompareGE(CaseValue value, const std::string& target)
    {
        if(value == GetMinValue())
            { EmitJump(target); return; }
        else if(value == GetMaxValue())
            EmitCompareEQ(value, target);
        else
         EmitCompareJump(value, target, "bcs");
    }
    virtual void EmitCompareLE(CaseValue value, const std::string& target)
    {
        if(value == GetMaxValue())
            { EmitJump(target); return; }
        else if(value == GetMinValue())
            EmitCompareEQ(value, target);
        else if(ChainCompare && value == LastCompare && LastCompareType == "beq")
            EmitCompareLT(value, target);
        else
            EmitCompareLT(value+1, target);
    }
    virtual void EmitCompareGT(CaseValue value, const std::string& target)
    {
        if(value == GetMaxValue())
            { /* nothing */ }
        else if(ChainCompare && value == LastCompare && LastCompareType == "beq")
            EmitCompareGE(value, target);
        else
            EmitCompareGE(value+1, target);
    }
    virtual void EmitSubtract(CaseValue value)
    {
        EmitImmed("sec:sbc", value);
        LastWasBra   = false;
    }
    virtual const std::string EmitBlock()
    {
        return GenLabel();
    }
    virtual void EmitEndBlock(const std::string& label)
    {
        ChainCompare = false;
        EmitLabel(label.c_str());
        LastWasBra   = false;
    }
    virtual void EmitJumpTable(const std::vector<std::string>& table)
    {
        Emit("asl:tax:jmp (+,x)");
        Emit("+");
        std::string line;
        for(unsigned a=0; a<table.size(); ++a)
        {
            if(line.empty()) line += ".word ";else line += ", ";
            line += table[a];
            if(line.size() > 60) { Emit(line.c_str()); line = ""; }
        }
        if(!line.empty()) Emit(line.c_str());
        LastWasBra   = true;
    }
};

/************ MAIN PROGRAM ***************/
static void AddList(CaseItemList& list, const char *set, const char *label)
{
    CaseItem tmp;
    tmp.target = label;
    while(*set) tmp.values.insert((unsigned char)*set++);
    list.push_back(tmp);
}
int main(void)
{
    CaseItemList list;
    
    /* Produce some test cases*/
    
    AddList(list, "aeiouy", "vowel");
    AddList(list, "bcdfghjklmnpqrstvwxz", "consonant");
    
    CaseHandler tmp;
    
    tmp.Generate(list, "nothing");
    
    return 0;
}
